/*
    ExInpaint Filter for Avisynth 2.5 -  Exemplar-Based Inpainting
	Copyright (C) 2008 A.G.Balakhnin aka Fizick, http://avisynth.org.ru

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "windows.h"
#include "avisynth.h"
#include "inpainting.h"

//-------------------------------------------------------------------------------------------
class ExInpaint : public GenericVideoFilter {

	//  parameters
	PClip maskclip;
	int color;
	int xsize;
	int ysize;
	int radius;
	int maxsteps;

	inpainting *inp;

public:

	ExInpaint(PClip _child,  PClip _maskclip, int _color, int _xsize, int _ysize, int _radius, int _maxsteps, IScriptEnvironment* env);
  ~ExInpaint();
	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
};


//Here is the acutal constructor code used
ExInpaint::ExInpaint(PClip _child, PClip _maskclip, int _color, int _xsize, int _ysize, int _radius, int _maxsteps, IScriptEnvironment* env):

	GenericVideoFilter(_child), maskclip(_maskclip), color(_color), xsize(_xsize), ysize(_ysize), radius(_radius), maxsteps(_maxsteps) {
  // This is the implementation of the constructor.
  // The child clip (source clip) is inherited by the GenericVideoFilter,
  //  where the following variables gets defined:
  //   PClip child;   // Contains the source clip.
  //   VideoInfo vi;  // Contains videoinfo on the source clip.

	int pixel_format;

    if (maskclip == 0) // no mask clip
    {
        if ( vi.IsRGB32() )
            pixel_format = RGBA; // use Alpha channel as a mask
        else
            env->ThrowError("ExInpaint: without mask clip video must be RGB32!");
    }
    else // have mask clip
    {
        if ( vi.IsRGB32() )
            pixel_format = RGB32; // but do not use Alpha channel
        else if ( vi.IsRGB24() )
            pixel_format = RGB24;
        else if ( vi.IsYV12() )
            pixel_format = YV12;
        else
            env->ThrowError("ExInpaint: video must be RGB32 or RGB24 or YV12!");

        VideoInfo maskvi = maskclip->GetVideoInfo();

        if (vi.width != maskvi.width || vi.height != maskvi.height )
            env->ThrowError("ExInpaint: mask size %d x %d is not as source clip size",maskvi.width,maskvi.height);

        if (vi.pixel_type != maskvi.pixel_type)
            env->ThrowError("ExInpaint: mask pixel type must be same as source clip type!");
    }

//	masklast = maskvi.num_frames;

// warning:
//    if (!(xsize%2 && ysize%2))
//		env->ThrowError("ExInpaint: xsize, ysize must be odd (3,5,7,9...)!");

	inp = new inpainting(vi.width, vi.height, pixel_format);

}
//-------------------------------------------------------------------------------------------

// This is where any actual destructor code used goes
ExInpaint::~ExInpaint() {
	delete inp;
}
//-------------------------------------------------------------------------------------------

PVideoFrame __stdcall ExInpaint::GetFrame(int n, IScriptEnvironment* env) {

    PVideoFrame maskframe;
	if (maskclip)
        maskframe = maskclip->GetFrame(n, env); // mask frame must be first to avoid makewritable bug

	PVideoFrame src = child->GetFrame(n, env);// Request frame 'n' from the child (source) clip.

	env->MakeWritable(&src); // will get results inplace

	int steps = 0;

	if (vi.IsYV12())
	{
		// This code deals with YV12 colourspace where the Y, U and V information are
		// stored in completely separate memory areas

		steps = inp->process3planes(src->GetWritePtr(PLANAR_Y),  src->GetPitch(PLANAR_Y),
			src->GetWritePtr(PLANAR_U), src->GetPitch(PLANAR_U),
			src->GetWritePtr(PLANAR_V),
			maskframe->GetReadPtr(PLANAR_Y), maskframe->GetPitch(PLANAR_Y),
			maskframe->GetReadPtr(PLANAR_U), maskframe->GetPitch(PLANAR_U),
			maskframe->GetReadPtr(PLANAR_V),
			xsize, ysize, radius, color, maxsteps); // inpaint frame

	}
	else if ( vi.IsRGB24() || (vi.IsRGB32() && maskclip!=0) )
	{

		steps = inp->process(src->GetWritePtr(),  src->GetPitch(),
			maskframe->GetReadPtr(), maskframe->GetPitch(),
			xsize, ysize, radius, color, maxsteps); // inpaint frame

	}
	else if (vi.IsRGB32() && maskclip==0)
	{

		steps = inp->process(src->GetWritePtr(),  src->GetPitch(),
			0, 0,
			xsize, ysize, radius, color, maxsteps); // inpaint frame

	}

	char buf[80];
	wsprintf(buf,"ExInpaint: frame=%d, steps=%d", n, steps);
	OutputDebugString(buf);

  // As we now are finished processing the image, we return the destination image.
	return src;
}

//-------------------------------------------------------------------------------------------

// This is the function that created the filter, when the filter has been called.
// This can be used for simple parameter checking, so it is possible to create different filters,
// based on the arguments recieved.

AVSValue __cdecl Create_ExInpaint(AVSValue args, void* user_data, IScriptEnvironment* env) {
    return new ExInpaint(args[0].AsClip(), // the 0th parameter is the source clip
		args[1].Defined() ? args[1].AsClip() : 0, // parameter mask
		 args[2].AsInt(0xFFFFFF), // parameter color of mask
		 args[3].AsInt(9), // parameter size
		 args[4].AsInt(9), // parameter size
		 args[5].AsInt(36), // parameter search radius
		 args[6].AsInt(100000), // parameter max steps
		 env);
    // Calls the constructor with the arguments provied.
}

//-------------------------------------------------------------------------------------------

// The following function is the function that actually registers the filter in AviSynth
// It is called automatically, when the plugin is loaded to see which functions this filter contains.

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env) {
    env->AddFunction("ExInpaint", "c[mask]c[color]i[xsize]i[ysize]i[radius]i[steps]i", Create_ExInpaint, 0);
    // The AddFunction has the following parameters:
    // AddFunction(Filtername , Arguments, Function to call,0);

    // Arguments is a string that defines the types and optional names of the arguments for you filter.
    // c - Video Clip
    // i - Integer number
    // f - Float number
    // s - String
    // b - boolean


    return "ExInpaint plugin";

}

